from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import Optional, List, Tuple, Union
import re

from .with_properties import WithProperties
from .with_version import WithVersion
from .with_files_metadata import WithFilesMetadata
from .files_metadata import TreeItemType, TreeItem
from .data_resource_schema import DataResourceSchema, SchemaColumn, DataType
from .artifact import _Base


NOTEBOOK = {"ipynb": True}
IMAGE_FILE = {"png": True, "jpeg": True, "svg": True}


def schema_validation(description, max_length, precision, scale):
    return (
        description if description is not None else 0,
        max_length if max_length is not None else 0,
        precision if precision is not None else 0,
        scale if scale is not None else 0,
    )


def extract_table_data(table, table_name):
    schema_columns, type_check = [], None
    for schema in table.schema:
        data_type = str(schema.field_type).lower().capitalize()
        try:
            type_check = DataType[data_type].__dict__["_value_"]
        except KeyError:
            pass
        schema_description, schema_max_length, schema_precision, schema_scale = schema_validation(
            schema.description, schema.max_length, schema.precision, schema.scale
        )
        schema_columns += [
            SchemaColumn(
                name=schema.name,
                description=schema_description,
                dataType=type_check,
                length=schema_max_length,
                precision=schema_precision,
                scale=schema_scale,
            )
        ]
    data_resource_schema = DataResourceSchema(
        type=TreeItemType.DataTable,
        name=table_name,
        description="",
        fileFormat="bigquery#table",
        columns=schema_columns,
    )
    return data_resource_schema


def get_all_files_in_folder(client, database_name, project_id) -> Tuple[List[TreeItem], int]:
    from google.cloud.bigquery import DatasetReference

    files_size = 0
    children: List[TreeItem] = []
    tables = client.list_tables(database_name)

    for curr_table in tables:
        table_name = curr_table.table_id
        table_reference = DatasetReference(project_id, database_name).table(table_name)
        table = client.get_table(table_reference)
        data_resource_schema = extract_table_data(table, table_name)
        table_name, created, updated, size = table.table_id, table.created, table.modified, table.num_bytes
        files_size += size
        uri = f"bigquery://{project_id}/bigquery-public-data.{database_name}.{table_name}"
        children += [
            TreeItem(
                name=table_name,
                uri=uri,
                itemCreatedDate=created,
                itemUpdatedDate=updated,
                size=size,
                type=TreeItemType.DataTable,
                metadata=data_resource_schema,
            )
        ]

    return children, files_size


def extract_bigquery_metadata(uri: str) -> Optional[DatasetMetadata]:
    from google.cloud import bigquery
    from google.cloud.bigquery import DatasetReference

    try:
        match_table, match_dataset = re.search(r"&d=(.*?)&p=(.*?)&t=(.*?)&page", uri), re.search(
            r"&d=(.*?)&p=(.*?)&page", uri
        )
        if match_table:
            database_name, project_id, table_name = match_table.group(1), match_table.group(2), match_table.group(3)
            client = bigquery.Client(project=project_id)
            table_reference = DatasetReference(project_id, database_name).table(table_name)
            table = client.get_table(table_reference)
            data_resource_schema = extract_table_data(table, table_name)
            name, created, updated, size = table.table_id, table.created, table.modified, table.num_bytes
            uri = f"bigquery://{project_id}/bigquery-public-data.{database_name}.{table_name}"
            return DatasetMetadata().with_metadata(
                name=name,
                uri=uri,
                itemCreatedDate=created,
                itemUpdatedDate=updated,
                size=size,
                type=TreeItemType.DataTable,
                metadata=data_resource_schema,
            )
        elif match_dataset:
            database_name, project_id = match_dataset.group(1), match_dataset.group(2)
            client = bigquery.Client(project=project_id)
            children, files_size = get_all_files_in_folder(client, database_name, project_id)
            uri = f"bigquery://{database_name}"
            return DatasetMetadata().with_metadata(
                name=database_name, uri=uri, size=files_size, isFolder=True, children=children, type=TreeItemType.Folder
            )
    except Exception as e:
        raise RuntimeError(f"Big Query Dataset creation failed due to {e}")
    return None


# Searches entire tree *NB the same folder names cause clashes.
def search_tree(node, search):
    if node.name == search:
        return node
    elif node.children:
        for child in node.children:
            result = search_tree(child, search)
            if result:
                return result
    # Could not find the result in any of the children
    return None


# searches children instead of the entire tree -> No clashes with similar folder names and stays at correct depth of the tree
def search_tree_children(node, search):
    if node.name == search:
        return node
    elif node.children:
        for child in node.children:
            if child.name == search:
                return child
    # Could not find the result in any of the children
    return None


def decode_hash(blob):
    import base64
    import binascii

    # decode the hash provided
    base64_message = blob.md5_hash
    md5_hash = binascii.hexlify(base64.urlsafe_b64decode(base64_message))
    return md5_hash.decode("utf-8")


def get_file_type(file_name):
    file_type = file_name.split(".", 1)
    if "csv" == file_type[1].lower():
        return "CsvFile"
    elif NOTEBOOK.get(file_type[1].lower()):
        return "Notebook"
    elif IMAGE_FILE.get(file_type[1].lower()):
        return "ImageFile"
    elif "md" == file_type[1].lower():
        return "MdFile"
    else:
        return "File"


def attach_queue(branch, trunk, curr_blob=None, full_path=None, bucket_name=None):
    """
    Insert a branch of directories on its trunk. Trunk is the TreeItem
    """
    # splits dict to current file path and blob for metadata
    if isinstance(branch, dict):
        curr_blob = branch["blob"]
        parts = branch["name"].split("/", 1)
        full_path = f"{bucket_name}/{parts[0]}"
    else:
        parts = branch.split("/", 1)
        if full_path is not None:
            full_path = f"{full_path}/{parts[0]}"
    # splits file paths into partitions of 1
    if len(parts) == 1:  # branch is a file
        # if a single file is provided e.g file.csv // double check this
        if trunk.name is None:
            file_type = get_file_type(parts[0])
            (
                trunk.name,
                trunk.isFolder,
                trunk.digest,
                trunk.path,
                trunk.type,
                trunk.size,
                trunk.uri,
                trunk.itemCreatedDate,
            ) = (
                parts[0],
                False,
                decode_hash(curr_blob),
                f"gs://{full_path}",
                file_type,
                curr_blob.size,
                curr_blob.name,
                curr_blob.time_created,
            )
            return trunk
        # catches any unintentional blob errors e.g a file path with no file
        elif len(parts[0]) <= 0:
            return trunk
        file_type = get_file_type(parts[0])
        trunk.children += [
            TreeItem(
                name=parts[0],
                isFolder=False,
                digest=decode_hash(curr_blob),
                uri=f"gs://{bucket_name}/{curr_blob.name}",
                type=file_type,
                size=curr_blob.size,
                itemCreatedDate=curr_blob.time_created,
            )
        ]
    else:
        node, others = parts
        # Add the root information needed for the search
        if trunk.name is None:
            trunk.name, trunk.type, trunk.uri, trunk.isFolder = node, "Folder", f"gs://{full_path}", True
        # search tree if node is present
        search = search_tree_children(trunk, node)
        if search is not None:
            trunk = search
        elif search is None:
            # TODO Appends to top TreeItem for same level of directory e.g root is the last point of reference for the search // *Tree and not Forest*
            trunk.append(
                TreeItem(
                    name=node,
                    digest=decode_hash(curr_blob),
                    path=curr_blob.self_link,
                    isFolder=True,
                    type="Folder",
                    uri=f"gs://{full_path}",
                )
            )
            trunk = search_tree_children(trunk, node)
        # if it isn't present, continue adding sub folders
        attach_queue(others, trunk, curr_blob, full_path, bucket_name)


def extract_gcs_metadata(uri: Union[str, List[str]]) -> Optional[DatasetMetadata]:
    from google.cloud import storage

    storage_client = storage.Client()
    # Use a Forest for a list of uris
    forest = []
    if isinstance(uri, str):
        uri = [uri]

    for item in uri:

        match = re.search("(.*?)/(.*)", item)
        if match is None:
            logging.warning(f"Unable to parse {item}. Please ensure it follows the guidelines.")
            return None
        bucket_name, blob = match.group(1), match.group(2)

        # keep blobs so a single query is only used
        blobs_query = list(storage_client.list_blobs(bucket_name, prefix=blob))
        blobs = [{"name": blob.name, "blob": blob} for blob in blobs_query]
        if len(blobs) <= 0:
            raise ValueError(f"The file/folder '{item}' could not be found please check the uri.")
        # Assign a tree
        tree = TreeItem()

        for line in blobs:
            # catch files to avoid building directories
            if line["name"] == blob:

                parts = line["name"].split("/")
                file_type = get_file_type(parts[-1])
                (
                    tree.name,
                    tree.isFolder,
                    tree.digest,
                    tree.path,
                    tree.type,
                    tree.size,
                    tree.uri,
                    tree.itemCreatedDate,
                ) = (
                    parts[-1],
                    False,
                    decode_hash(line["blob"]),
                    f"gs://{line['name']}",
                    file_type,
                    line["blob"].size,
                    line["blob"].name,
                    line["blob"].time_created,
                )
            else:
                attach_queue(line, tree, bucket_name=bucket_name)
        forest += [tree]
    dataset = DatasetMetadata().with_metadata()
    dataset.filesMetadata = forest
    return dataset


# TODO Python 3.10 kw_only=true -> will allow for more flexible dataclass inheritance
@dataclass
class DatasetMetadata(_Base, WithVersion, WithProperties, WithFilesMetadata):
    autoVersion: bool = True

    @classmethod
    def create_bigquery(cls, uri: str) -> Optional[DatasetMetadata]:
        dataset_metadata_artifact = extract_bigquery_metadata(uri)
        return dataset_metadata_artifact

    @classmethod
    def create_gcs(cls, uri: Union[str, List[str]]) -> Optional[DatasetMetadata]:
        dataset_metadata_artifact = extract_gcs_metadata(uri)
        return dataset_metadata_artifact



from enum import EnumMeta


class JobArtifactType(EnumMeta):
    """
    Indicates whether the artifact is an input or an output of a run
    """

    INPUT = "INPUT"
    """
    """
    OUTPUT = "OUTPUT"
    """
    """



from typing import List, Optional, Tuple, Any, BinaryIO

from vectice.api.Page import Page
from vectice.api.attachment import AttachmentApi
from vectice.api.connections import ConnectionApi
from vectice.api.dataset import DatasetApi
from vectice.api.dataset_version import DatasetVersionApi
from vectice.api.job import JobApi
from vectice.api.job_artifact import ArtifactApi
from vectice.api.job_run import RunApi
from vectice.api.json_object import JsonObject
from vectice.api.model import ModelApi
from vectice.api.model_version import ModelVersionApi
from vectice.api.output.attachments_output import AttachmentOutput
from vectice.api.output.connection_output import ConnectionOutput
from vectice.api.output.dataset_output import DatasetOutput
from vectice.api.output.dataset_version_output import DatasetVersionOutput
from vectice.api.output.job_output import JobOutput
from vectice.api.output.job_run_output import JobRunOutput
from vectice.api.output.metric_output import EntityMetricOutput
from vectice.api.output.model_output import ModelOutput
from vectice.api.output.model_version_output import ModelVersionOutput
from vectice.api.output.paged_response import PagedResponse
from vectice.api.output.property_output import EntityPropertyOutput
from vectice.api.project import ProjectApi
from vectice.api.rule import RuleApi
from vectice.models import Artifact, RunnableJob


class Client(ProjectApi):
    """
    Low level Vectice API client.
    """

    def __init__(self, project_token: str, auto_connect=True, allow_self_certificate=True):
        super().__init__(
            project_token=project_token, auto_connect=auto_connect, allow_self_certificate=allow_self_certificate
        )
        self._allow_self_certificate = allow_self_certificate
        self._api_config = {
            "project_token": self.project_token,
            "allow_self_certificate": self._allow_self_certificate,
            "_token": self._token,
        }

    @property
    def _config(self) -> dict:
        self._api_config["_token"] = self._token
        self._api_config["allow_self_certificate"] = self._allow_self_certificate
        self._api_config["project_token"] = self.project_token
        return self._api_config

    def start_run(
        self,
        run: RunnableJob,
        inputs: Optional[List[Artifact]] = None,
    ) -> JsonObject:
        """
        Start a run.

        :param run: The run to start
        :param inputs: A list of artifacts to linked to the run
        :return: A json object
        """
        return RuleApi(**self._config).start_run(run.job, run.run, inputs)

    def stop_run(self, run: JsonObject, outputs: Optional[List[Artifact]] = None):
        """

        :param run:
        :param outputs:
        :return:
        """
        return RuleApi(**self._config).stop_run(run, outputs)

    def list_jobs(
        self, search: Optional[str] = None, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[JobOutput]:
        """

        :param search: A text to filter jobs we are looking for
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return: A paged response that contains a list of JobOutput instances.
        """
        return JobApi(**self._config).list_jobs(search, page_index, page_size)

    def create_job(self, job: JsonObject) -> JobOutput:
        """
        Create a job

        :param job: A job description (json)
        :return: A JobOutput instance
        """
        return JobApi(**self._config).create_job(job)

    def update_job(self, job_id: int, job: JsonObject):
        """
        Update a job

        :param job: A job description (json)
        :return: The json structure
        """
        return JobApi(**self._config).update_job(job_id, job)

    def list_runs(self, job_id: int, page_index=Page.index, page_size=Page.size) -> List[JobRunOutput]:
        """
        List runs of a specific job.

        :param job_id: The identifier of the job
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return: a list of JobRunOutput
        """
        return RunApi(job_id=job_id, **self._config).list_runs(page_index, page_size)

    def create_run(self, job_id: int, run: JsonObject):
        """
        Create a run

        :param job_id: The identifier of the job
        :param run: A run description (json)
        :return: The json structure
        """
        return RunApi(job_id=job_id, **self._config).create_run(run)

    def update_run(self, job_id: int, run_id: int, run: JsonObject):
        """
        Update a run

        :param job_id: The identifier of the job
        :param run_id:
        :param run:
        :return: The json structure
        """
        return RunApi(job_id=job_id, **self._config).update_run(run_id, run)

    def create_artifact(self, job_id: int, run_id: int, artifact: JsonObject):
        """
        Create artifact

        :param job_id: The identifier of the job
        :param run_id:
        :param artifact:
        :return: The json structure
        """
        return ArtifactApi(job_id=job_id, run_id=run_id, **self._config).create_artifact(artifact)

    def update_artifact(self, job_id: int, run_id: int, artifact_id: int, artifact: JsonObject):
        """
        Update artifact

        :param job_id: The identifier of the job
        :param run_id:
        :param artifact_id:
        :param artifact:
        :return: The json structure
        """
        return ArtifactApi(job_id=job_id, run_id=run_id, **self._config).update_artifact(artifact_id, artifact)

    def list_datasets(
        self, search: str = None, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[DatasetOutput]:
        """
        List datasets

        :param search: a text used to filter list of datasets
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return: The json structure
        """
        return DatasetApi(**self._config).list_datasets(search, page_index, page_size)

    def create_dataset(self, dataset: JsonObject):
        """
        Create a dataset

        :param dataset:
        :return: The json structure
        """
        return DatasetApi(**self._config).create_dataset(dataset)

    def update_dataset(self, dataset_id: int, dataset: JsonObject):
        """
        Update a dataset

        :param dataset_id:
        :param dataset:
        :return: The json structure
        """
        return DatasetApi(**self._config).update_dataset(dataset_id, dataset)

    def list_models(self, search: str = None, page_index=Page.index, page_size=Page.size) -> PagedResponse[ModelOutput]:
        """
        List models

        :param search:
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return: The json structure
        """
        return ModelApi(**self._config).list_models(search, page_index, page_size)

    def create_model(self, model: JsonObject):
        """
        Create a model

        :param model:
        :return: The json structure
        """
        return ModelApi(**self._config).create_model(model)

    def update_model(self, model_id: int, model: JsonObject):
        """
        Update a model

        :param model_id:
        :param model:
        :return: The json structure
        """
        return ModelApi(**self._config).update_model(model_id, model)

    def list_dataset_versions(
        self, dataset_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[DatasetVersionOutput]:
        """
        List dataset versions

        :param dataset_id:
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return:
        """
        return DatasetVersionApi(dataset_id=dataset_id, **self._config).list_dataset_versions(page_index, page_size)

    def create_dataset_version(self, dataset_id: int, dataset_version: JsonObject):
        """
        Create a dataset version

        :param dataset_id:
        :param dataset_version:
        :return: The json structure
        """
        return DatasetVersionApi(dataset_id=dataset_id, **self._config).create_dataset_version(dataset_version)

    def update_dataset_version(self, dataset_id: int, dataset_version_id: int, dataset_version: JsonObject):
        """
        Update a dataset version

        :param dataset_id:
        :param dataset_version_id:
        :param dataset_version:
        :return: The json structure
        """
        return DatasetVersionApi(dataset_id=dataset_id, **self._config).update_dataset_version(
            dataset_version_id, dataset_version
        )

    def get_dataset_version(
        self, dataset_id: int, search: str, page_index=Page.index, page_size=Page.size
    ) -> Optional[DatasetVersionOutput]:
        """
        Get a dataset version

        :param dataset_id:
        :param search:
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return:
        """
        return DatasetVersionApi(dataset_id=dataset_id, **self._config).get_dataset_version(
            search, page_index, page_size
        )

    def create_dataset_version_properties(self, dataset_id: int, dataset_version_id: int, properties: List[dict]):
        """
        Create data version properties

        :param dataset_id:
        :param dataset_version_id:
        :param properties:
        :return:
        """
        return DatasetVersionApi(dataset_id=dataset_id, **self._config).create_dataset_version_properties(
            dataset_version_id, properties
        )

    def update_dataset_version_properties(
        self, dataset_id: int, dataset_version_id: int, property_id: int, properties: JsonObject
    ):
        """
        Create data version properties

        :param dataset_id:
        :param dataset_version_id:
        :param properties:
        :param property_id:
        :return:
        """
        return DatasetVersionApi(dataset_id=dataset_id, **self._config).update_dataset_version_properties(
            dataset_version_id, property_id, properties
        )

    def list_dataset_version_properties(
        self, dataset_id: int, dataset_version_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[EntityPropertyOutput]:
        """
        List dataset version properties

        :param dataset_id:
        :param dataset_version_id:
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return:
        """
        return DatasetVersionApi(dataset_id=dataset_id, **self._config).list_dataset_version_properties(
            dataset_version_id, page_index, page_size
        )

    def list_model_versions(
        self, model_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[ModelVersionOutput]:
        """
        List model versions

        :param model_id:
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return:
        """
        return ModelVersionApi(model_id=model_id, **self._config).list_model_versions(page_index, page_size)

    def list_model_version_metrics(
        self, model_id: int, model_version_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[EntityMetricOutput]:
        """
        List model versions

        :param model_id:
        :param model_version_id:
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return:
        """
        return ModelVersionApi(model_id=model_id, **self._config).list_model_version_metrics(
            model_version_id, page_index, page_size
        )

    def list_model_version_properties(
        self, model_id: int, model_version_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[EntityPropertyOutput]:
        """
        List model versions

        :param model_id:
        :param model_version_id:
        :param page_index: The index of the page
        :param page_size: The size of the page
        :return:
        """
        return ModelVersionApi(model_id=model_id, **self._config).list_model_version_properties(
            model_version_id, page_index, page_size
        )

    def create_model_version_properties(self, model_id: int, model_version_id: int, properties: List[dict]):
        """
        Create model version properties

        :param model_id:
        :param model_version_id:
        :param properties:
        :return:
        """
        return ModelVersionApi(model_id=model_id, **self._config).create_model_version_properties(
            model_version_id, properties
        )

    def update_model_version_properties(
        self, model_id: int, model_version_id: int, property_id: int, properties: JsonObject
    ):
        """
        Create model version properties

        :param model_id:
        :param model_version_id:
        :param properties:
        :param property_id:
        :return:
        """
        return ModelVersionApi(model_id=model_id, **self._config).update_model_version_properties(
            model_version_id, property_id, properties
        )

    def create_model_version(self, model_id: int, model_version: JsonObject):
        """
        Create a model version

        :param model_id:
        :param model_version:
        :return: The json structure
        """
        return ModelVersionApi(model_id=model_id, **self._config).create_model_version(model_version)

    def update_model_version(self, model_id: int, model_version_id: int, model_version: JsonObject):
        """
        Update a model version

        :param model_id:
        :param model_version_id:
        :param model_version:
        :return: The json structure
        """
        return ModelVersionApi(model_id=model_id, **self._config).update_model_version(model_version_id, model_version)

    def get_attachment(self, artifact_type: str, artifact_version: int, file_id: int):
        """
        Downloads the specified attachment

        :param artifact_type:
        :param artifact_version:
        :param file_id:
        :return: The file requested
        """
        return AttachmentApi(**self._config).get_attachment(artifact_type, artifact_version, file_id)

    def create_attachments(
        self, artifact_type: str, artifact_version: int, files: Optional[List[Tuple[str, Tuple[Any, BinaryIO]]]]
    ):
        """
        Create an attachment

        :param artifact_type:
        :param artifact_version:
        :param files:
        :return: The json structure
        """
        return AttachmentApi(**self._config).post_attachment(artifact_type, artifact_version, files)

    def update_attachments(
        self, artifact_type: str, artifact_version: int, files: List[Tuple[str, Tuple[Any, BinaryIO]]]
    ):
        """
        Update an attachment

        :param artifact_type:
        :param artifact_version:
        :param files:
        :return: The json structure
        """
        return AttachmentApi(**self._config).update_attachments(artifact_type, artifact_version, files)

    def delete_attachment(self, artifact_type: str, artifact_version: int, file_id: int):
        """
        Delete attachment

        :param artifact_type:
        :param artifact_version:
        :param file_id:
        :return: The json structure
        """
        return AttachmentApi(**self._config).delete_attachment(artifact_type, artifact_version, file_id)

    def list_attachments(self, artifact_type: str, artifact_version: int) -> PagedResponse[AttachmentOutput]:
        """
        List attachments

        :param artifact_type:
        :param artifact_version:
        :return: The json structure
        """
        return AttachmentApi(**self._config).list_attachments(artifact_type, artifact_version)

    def list_connections(
        self, search_name: Optional[str] = None, connection_type: Optional[str] = None
    ) -> PagedResponse[ConnectionOutput]:
        """
        List connections

        :param search_name:
        :param connection_type:
        :return: The json structure
        """
        api = ConnectionApi(self._config["_token"], allow_self_certificate=self._config["allow_self_certificate"])
        return api.list_connections(search_name=search_name, connection_type=connection_type)



from __future__ import annotations

from dataclasses import dataclass, asdict
from typing import Optional, List, Union, Tuple
from .with_properties import WithProperties
from .with_tags import WithTags


@dataclass
class CreateDataset(WithProperties, WithTags):
    name: Optional[str] = None
    connectionId: Optional[int] = None
    connectionName: Optional[str] = None
    notes: Optional[str] = None
    files: Optional[List[str]] = None
    folders: Optional[List[str]] = None
    data_properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None
    data_tags: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None

    def _create_with_property(self, data_properties: Tuple[str, str]):
        key, value = data_properties
        return self.with_property(key, value)

    def _create_with_properties(self, data_properties: List[Tuple[str, str]]):
        return self.with_properties(properties=data_properties)

    def add_properties(
        self, data_properties: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None
    ) -> CreateDataset:
        if data_properties and isinstance(data_properties, list):
            self._create_with_properties(data_properties=data_properties)
        elif data_properties and isinstance(data_properties, tuple):
            self._create_with_property(data_properties=data_properties)
        return self

    def _create_with_tag(self, data_tags: Tuple[str, str]):
        key, value = data_tags
        return self.with_tag(key, value)

    def _create_with_tags(self, data_tags: List[Tuple[str, str]]):
        return self.with_tags(tags=data_tags)

    def add_tags(self, data_tags: Optional[Union[List[Tuple[str, str]], Tuple[str, str]]] = None) -> CreateDataset:
        if data_tags and isinstance(data_tags, list):
            self._create_with_tags(data_tags=data_tags)
        elif data_tags and isinstance(data_tags, tuple):
            self._create_with_tag(data_tags=data_tags)
        return self

    def as_dict(self):
        dataset = asdict(self)
        dataset_clean = dataset.copy()
        for key, value in dataset.items():
            if value is None:
                del dataset_clean[key]
            if key == "notes" and value is not None:
                dataset_clean["description"] = dataset_clean.pop("notes")
        return dataset_clean



from typing import Optional

from vectice.api._auth import Auth


class ProjectApi(Auth):
    def __init__(
        self, project_token: str, _token: Optional[str] = None, auto_connect=True, allow_self_certificate=True
    ):
        super().__init__(_token=_token, auto_connect=auto_connect, allow_self_certificate=allow_self_certificate)
        self._project_token = project_token
        self._project_path = "/metadata/project/" + str(project_token)

    @property
    def project_token(self) -> str:
        return self._project_token

    @property
    def api_base_path(self) -> str:
        return self._project_path



from .dataset_version_output import DatasetVersionOutput
from .dataset_output import DatasetOutput
from .model_output import ModelOutput
from .model_version_output import ModelVersionOutput
from .job_output import JobOutput
from .job_run_output import JobRunOutput
from .attachments_output import AttachmentOutput
from .paged_response import PagedResponse

__all__ = [
    "DatasetOutput",
    "DatasetVersionOutput",
    "JobOutput",
    "JobRunOutput",
    "ModelOutput",
    "ModelVersionOutput",
    "AttachmentOutput",
    "PagedResponse",
]



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TypeVar, List, Optional

from .data_resource import DataResource


class WithDataResourcesTrait(ABC):
    T = TypeVar("T", bound="WithDataResourcesTrait")

    @abstractmethod
    def with_resources(self: T, files: Optional[List[str]] = None, folders: Optional[List[str]] = None) -> T:
        """
        :param files: the files to put in the dataset
        :param folders: the folders (and all file in it) to put in the dataset
        :return: itself
        """
        pass


@dataclass
class WithDataResources(WithDataResourcesTrait):
    dataResources: Optional[List[DataResource]] = None

    T = TypeVar("T", bound="WithDataResources")

    def with_resources(self: T, files: Optional[List[str]] = None, folders: Optional[List[str]] = None) -> T:
        if folders is None:
            folders = []
        if files is None:
            files = []
        file_resources = [DataResource.create_item_resource(uri) for uri in files]
        folder_resources = [DataResource.create_folder_resource(uri) for uri in folders]
        self.dataResources = file_resources + folder_resources
        return self


class WithDelegatedDataResources(WithDataResourcesTrait, ABC):
    T = TypeVar("T", bound="WithDelegatedDataResources")

    @abstractmethod
    def _get_delegate(self) -> WithDataResourcesTrait:
        pass

    def with_resources(self: T, files: Optional[List[str]] = None, folders: Optional[List[str]] = None) -> T:
        self._get_delegate().with_resources(files, folders)
        return self



from dataclasses import dataclass
from typing import Optional


@dataclass
class UserVersion:
    description: Optional[str]
    """
    """
    name: Optional[str]
    """
    """
    targetId: Optional[int]
    """
    """
    userVersion: Optional[str]
    """
    """
    uri: Optional[str]
    """
    """



import logging
import re
import requests
import urllib3

from json import JSONEncoder, JSONDecoder
from requests import Response
from typing import Dict, Any, cast, Optional, List, Tuple, BinaryIO, Sequence
from vectice.api.json_object import JsonObject

DEFAULT_API_ENDPOINT = "http://localhost:4000"

logger = logging.getLogger("vectice.http")


class HttpError(Exception):
    def __init__(self, code: int, reason: str, path: str, method: str, json: Optional[str]):
        super().__init__()
        self.code: int = code
        self.reason: str = reason
        self.path = path
        self.method = method
        self.json = json

    def __str__(self):
        return f"""HTTP Error Code {self.code} : {self.reason}
        {self.method} {self.path}

        {self.json}
        """


def format_url(url: str) -> str:
    """Add https protocol if missing and remove trailing slash."""
    url = url.rstrip("/")
    if not re.match("(?:http|https|ftp)://", url):
        return "https://{}".format(url)
    return url


class VecticeEncoder(JSONEncoder):
    """
    Json Encoder with 2 specific behaviors:
    - handle datetime types so be serialized as a string following ISO8601 format
    - remove any null property from the serialized json.
    """

    def default(self, obj: Any) -> Any:
        if hasattr(obj, "isoformat"):
            return obj.isoformat()
        internal_copy = obj.__dict__.copy()
        return {k: v for (k, v) in internal_copy.items() if v is not None}


# DecodedType = TypeVar('DecodedType')


class Connection:
    def __init__(self, api_endpoint: Optional[str] = None, allow_self_certificate=True):
        self._API_BASE_URL = format_url(api_endpoint or DEFAULT_API_ENDPOINT)
        self._request_headers: Dict[str, str] = {}
        self.verify_certificate = not allow_self_certificate
        if allow_self_certificate:
            urllib3.disable_warnings()

    @property
    def api_base_url(self) -> str:
        return self._API_BASE_URL

    def _get(self, path: str) -> Dict[str, Any]:
        self._request_headers["Content-Type"] = "application/json"
        response = requests.get(
            url=self.api_base_url + path, headers=self._request_headers, verify=self.verify_certificate
        )
        return self._response(self.api_base_url + path, response, "GET")

    def _post(self, path: str, payload: Dict[str, Any] = None, decoder=JSONDecoder, files=None) -> JsonObject:
        self._request_headers["Content-Type"] = "application/json"
        data = VecticeEncoder(indent=1).encode(payload)
        response = requests.post(
            url=self.api_base_url + path, headers=self._request_headers, data=data, verify=self.verify_certificate
        )
        return self._response(self.api_base_url + path, response, "POST", payload)

    def _put(self, path: str, payload: Any = None, decoder=JSONDecoder, cls=dict) -> Dict[str, Any]:
        self._request_headers["Content-Type"] = "application/json"
        data = VecticeEncoder().encode(payload)
        response = requests.put(
            url=self.api_base_url + path, headers=self._request_headers, data=data, verify=self.verify_certificate
        )
        return self._response(self.api_base_url + path, response, "POST", payload)

    def _response(
        self, path: str, response: Response, method: str, payload: Optional[Dict[str, Any]] = None
    ) -> Dict[str, Any]:
        self.raise_status(path, response, method, payload)
        logger.debug(f"{method} {path} {response.status_code}")
        logger.debug("\n".join(f"{item[0]}: {item[1]}" for item in self._request_headers.items()))
        logger.debug(VecticeEncoder(indent=4, sort_keys=True).encode(payload))
        if len(response.content) > 0:
            return cast(Dict[str, Any], response.json())
        return {response.reason: response.status_code}

    @classmethod
    def raise_status(cls, path: str, response: Response, method: str, payload: Optional[Dict[str, Any]]) -> None:
        if not (200 <= response.status_code < 300):
            reason = response.text
            json = VecticeEncoder(indent=4, sort_keys=True).encode(payload) if payload is not None else None
            raise HttpError(response.status_code, reason, path, method, json)

    @classmethod
    def raise_attachment_status(cls, path: str, response: Response, method: str) -> None:
        if not (200 <= response.status_code < 300):
            reason = response.text
            raise ConnectionError(reason)

    def _post_attachments(
        self, path: str, files: Optional[List[Tuple[str, Tuple[Any, BinaryIO]]]] = None
    ) -> Optional[Response]:
        if self._request_headers.get("Content-Type"):
            self._request_headers.pop("Content-Type")
        response = requests.post(
            url=self.api_base_url + path,
            headers=self._request_headers,
            files=files,
            verify=self.verify_certificate,
        )
        return self._attachment_response(self.api_base_url + path, response, "POST")

    def _update_attachments(
        self, path: str, files: Optional[List[Tuple[str, Tuple[Any, BinaryIO]]]] = None
    ) -> Optional[Response]:
        if self._request_headers.get("Content-Type"):
            self._request_headers.pop("Content-Type")
        response = requests.put(
            url=self.api_base_url + path,
            headers=self._request_headers,
            files=files,
            verify=self.verify_certificate,
        )
        return self._attachment_response(self.api_base_url + path, response, "PUT")

    def _get_attachment(self, path: str) -> Optional[Any]:
        if self._request_headers.get("Content-Type"):
            self._request_headers.pop("Content-Type")
        response = requests.get(
            url=self.api_base_url + path, headers=self._request_headers, verify=self.verify_certificate
        )
        self._attachment_response(path=self.api_base_url + path, response=response, method="GET")
        return response

    def _delete_attachment(self, path: str) -> Optional[Response]:
        response = requests.delete(url=self.api_base_url + path, headers=self._request_headers)
        return self._attachment_response(self.api_base_url + path, response, "DELETE")

    def _list_attachments(self, path: str) -> Sequence[dict]:
        response = requests.get(
            url=self.api_base_url + path, headers=self._request_headers, verify=self.verify_certificate
        )
        self._attachment_response(self.api_base_url + path, response, "GET")
        return cast(Sequence[Dict], response.json())

    def _attachment_response(self, path: str, response: Response, method: str) -> Optional[Response]:
        self.raise_attachment_status(path, response, method)
        logger.debug(f"{method} {path} {response.status_code}")
        logger.debug("\n".join(f"{item[0]}: {item[1]}" for item in self._request_headers.items()))
        return response



from __future__ import annotations

from dataclasses import dataclass
from typing import List, Tuple

from .job import Job
from .job_run import JobRun


@dataclass
class RunnableJob:
    job: Job
    run: JobRun

    def with_property(self, key: str, value: str) -> RunnableJob:
        self.run.with_property(key, value)
        return self

    def with_properties(self, properties: List[Tuple[str, str]]) -> RunnableJob:
        self.run.with_properties(properties)
        return self

    def with_tag(self, key: str, value: str) -> RunnableJob:
        self.run.with_tag(key, value)
        return self



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, Tuple, List, Any, BinaryIO, TypeVar
from abc import ABC, abstractmethod


class WithAttachmentsTrait(ABC):
    T = TypeVar("T", bound="WithAttachmentsTrait")

    @abstractmethod
    def with_attachments(self, files: List[str]):
        """Accepts a list of file paths and then attaches the files to the modelversion that is created.
        e.g ["filepath/image.png", r"User/path/metrics.json"]

        :param files: The list of files
        :return: ModelVersionArtifact
        """
        pass


@dataclass
class WithAttachments(WithAttachmentsTrait):
    artifact_type: Optional[str] = None
    """"""
    files: Optional[List[Tuple[str, Tuple[Any, BinaryIO]]]] = None
    """"""

    def with_attachments(self, files: List[str]):
        self.files = [("file", (file, open(file, "rb"))) for file in files]
        return self


class WithDelegatedAttachments(WithAttachmentsTrait, ABC):
    T = TypeVar("T", bound="WithDelegatedAttachments")

    @abstractmethod
    def _get_attachment(self) -> WithAttachmentsTrait:
        pass

    def with_attachments(self: T, files: List[str]) -> T:
        self._get_attachment().with_attachments(files)
        return self



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional, List, Union
from enum import EnumMeta

from .data_resource_schema import DataResourceSchema


class TreeItemType(EnumMeta):
    Folder = "Folder"
    CsvFile = "CsvFile"
    ImageFile = "ImageFile"
    ExcelFile = "ExcelFile"
    TextFile = "TextFile"
    MdFile = "MdFile"
    DataSet = "DataSet"
    DataTable = "DataTable"
    File = "File"
    Notebook = "Notebook"


@dataclass
class TreeItem:
    name: Optional[str] = None
    id: Optional[str] = None
    parentId: Optional[str] = None
    path: Optional[str] = None
    type: Optional[Union[TreeItemType, str]] = None
    isFolder: Optional[bool] = False
    children: List[TreeItem] = field(default_factory=list)
    size: Optional[int] = 0
    uri: Optional[str] = None
    metadata: Optional[DataResourceSchema] = None
    digest: Optional[str] = None
    itemCreatedDate: Optional[str] = None
    itemUpdatedDate: Optional[str] = None

    def append(self, child):
        return self.children.append(child)



import logging

logger = logging.getLogger(__name__)


class LibraryFactory:
    @staticmethod
    def get_library(project_token: str, lib: str, *args, **kwargs):
        if str(lib).lower() == "mlflow":
            from .mlflow import MlflowAdapter

            return MlflowAdapter(project_token=project_token, *args, **kwargs)  # type: ignore
        else:
            raise ValueError(f"Unsupported library: {lib}")



from typing import Optional
from urllib.parse import urlencode

from vectice.entity import Job

from .Page import Page
from .json_object import JsonObject
from .output.job_output import JobOutput
from .output.paged_response import PagedResponse
from .project import ProjectApi


class JobApi(ProjectApi):
    def __init__(self, project_token: str, _token: Optional[str] = None, allow_self_certificate=True):
        super().__init__(project_token=project_token, _token=_token, allow_self_certificate=allow_self_certificate)
        self._job_path = super().api_base_path + "/job"

    @property
    def api_base_path(self) -> str:
        return self._job_path

    def list_jobs(self, search: str = None, page_index=Page.index, page_size=Page.size) -> PagedResponse[JobOutput]:
        queries = {"index": page_index, "size": page_size}
        if search:
            queries["search"] = search
        jobs = self._get(self.api_base_path + "?" + urlencode(queries))
        return PagedResponse(item_cls=JobOutput, total=jobs["total"], page=jobs["page"], items=jobs["items"])

    def create_job(self, job: JsonObject) -> JobOutput:
        if job.get("name") is None:
            raise ValueError('"name" must be provided in job.')
        if job.get("type") is None:
            job["type"] = "OTHER"
        return JobOutput(**self._post(self.api_base_path, job))

    def update_job(self, job_id: int, job: JsonObject) -> Job:
        return Job(self._put(self.api_base_path + "/" + str(job_id), job))



import sys
import os
from setuptools import setup

if sys.version_info < (3, 7):
    print("vectice requires Python 3 version >= 3.7.", file=sys.stderr)
    sys.exit(1)

package_root = os.path.abspath(os.path.dirname(__file__))

readme_filename = os.path.join(package_root, "README.rst")

version_requires = ">=3.7.1"

with open(readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read()
    readme.replace("Python >= 3.7.1", f"Python {version_requires}")


setup(
    name="vectice",
    version="0.23.1",
    description="Vectice Python library",
    long_description=readme,
    author="Vectice Inc.",
    author_email="sdk@vectice.com",
    url="https://github.com/vectice/vectice-python",
    packages=[
        "vectice",
        "vectice.api",
        "vectice.api.output",
        "vectice.adapter",
        "vectice.entity",
        "vectice.models",
        "vectice.utils",
    ],
    license="Apache License 2.0",
    keywords=["Vectice", "Client", "API", "Adapter"],
    platforms=["Linux", "MacOS X", "Windows"],
    python_requires=version_requires,
    install_requires=[
        "requests >= 2.5.0",
        "urllib3",
        "python-dotenv >= 0.10.0",
    ],
    tests_require=["mock >= 1.0.1", "pytest", "coverage", "pytest-cov", "testcontainers"],
    extras_require={
        "dev": ["bandit", "black", "flake8", "mypy", "pre-commit", "Pygments"],
        "mlflow": ["mlflow"],
        "doc": ["sphinx", "recommonmark", "nbsphinx", "sphinx-rtd-theme", "pypandoc", "jupyterlab"],
        "git": ["GitPython"],
        "github": ["PyGithub"],
        "bitbucket": ["atlassian-python-api"],
        "gitlab": ["python-gitlab"],
        "pandas": ["pandas"],
        "jupyter": ["jupyter_core", "ipykernel", "traitlets"],
        "collab": ["oauth2client", "GitPython >= 3.1.14"],
        "bigquery": ["google-cloud-bigquery >= 2.32.0", "google-cloud-bigquery-storage >= 2.11.0"],
        "gcs": ["google-cloud-storage"],
    },
    classifiers=[
        "Topic :: Internet",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Natural Language :: English",
        "Typing :: Typed",
    ],
)



from enum import Enum

from vectice.entity._versioned_entity import VersionedEntity


class Run(VersionedEntity):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def status(self) -> str:
        return str(self["status"])

    @property
    def job_id(self) -> int:
        return int(self["jobId"])

    @property
    def start_date(self) -> str:
        return str(self["startDate"])

    @property
    def end_date(self) -> str:
        return str(self["endDate"])

    @property
    def duration(self) -> int:
        return int(self["duration"])

    @property
    def system_name(self) -> str:
        return str(self["systemName"])

    @property
    def metadata_source(self) -> str:
        return str(self["metadataSource"])


class JobRunStatus(Enum):
    SCHEDULED = "SCHEDULED"
    STARTED = "STARTED"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"
    ABORTED = "ABORTED"



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional, TypeVar

from .artifact_version import ArtifactVersion
from .user_declared_version import UserDeclaredVersion


class WithVersionTrait(ABC):
    T = TypeVar("T", bound="WithVersionTrait")

    @abstractmethod
    def with_user_version(
        self: T,
        name: Optional[str] = None,
        uri: Optional[str] = None,
        description: Optional[str] = None,
    ) -> T:
        """
        create a new version from information given as parameters.

        :param name: the human name of the version
        :param uri: an uri to point to the version in the user system.
        :param description:
        :return: itself
        """
        pass

    @abstractmethod
    def with_existing_version_number(self: T, version_number: int) -> T:
        """
        :param version_number: the version number to use
        :return: itself
        """
        pass

    @abstractmethod
    def with_existing_version_name(self: T, version_name: str) -> T:
        """
        :param version_name: the version name to use
        :return: itself
        """
        pass

    @abstractmethod
    def with_existing_version_id(self: T, version_id: int) -> T:
        """
        :param version_id: the version id to use
        :return: itself
        """
        pass

    @abstractmethod
    def with_generated_version(self: T) -> T:
        """
        :return: itself
        """
        pass


@dataclass
class WithVersion(WithVersionTrait):
    version: Optional[ArtifactVersion] = None
    """
    existing version of the artifact defined by an id or a name
    """
    userDeclaredVersion: Optional[UserDeclaredVersion] = None
    """
    new version of the artifact defined by the user.

    The version can only be declared once.
    """

    T = TypeVar("T", bound="WithVersion")

    def with_user_version(
        self: T,
        name: Optional[str] = None,
        uri: Optional[str] = None,
        description: Optional[str] = None,
    ) -> T:
        self.userDeclaredVersion = UserDeclaredVersion(name, uri, description)
        self.version = None
        return self

    def with_existing_version_number(self: T, version_number: int) -> T:
        self.userDeclaredVersion = None
        self.version = ArtifactVersion(version_number, None, None)
        return self

    def with_existing_version_name(self: T, version_name: str) -> T:
        self.userDeclaredVersion = None
        self.version = ArtifactVersion(None, version_name, None)
        return self

    def with_existing_version_id(self: T, version_id: int) -> T:
        self.userDeclaredVersion = None
        self.version = ArtifactVersion(None, None, version_id)
        return self

    def with_generated_version(self: T):
        self.userDeclaredVersion = UserDeclaredVersion()
        return self


class WithDelegatedVersion(WithVersionTrait, ABC):
    T = TypeVar("T", bound="WithDelegatedVersion")

    @abstractmethod
    def _get_delegate(self) -> WithVersionTrait:
        pass

    def with_user_version(
        self: T,
        name: Optional[str] = None,
        uri: Optional[str] = None,
        description: Optional[str] = None,
    ) -> T:
        self._get_delegate().with_user_version(name, uri, description)
        return self

    def with_existing_version_number(self: T, version_number: int) -> T:
        self._get_delegate().with_existing_version_number(version_number)
        return self

    def with_existing_version_name(self: T, version_name: str) -> T:
        self._get_delegate().with_existing_version_name(version_name)
        return self

    def with_existing_version_id(self: T, version_id: int) -> T:
        """
        :param version_id: the version id to use
        :return: itself
        """
        self._get_delegate().with_existing_version_id(version_id)
        return self

    def with_generated_version(self: T):
        self._get_delegate().with_generated_version()
        return self



from .Page import Page
from .client import Client
from .dataset import DatasetApi
from .dataset_version import DatasetVersionApi
from .job import JobApi
from .job_artifact import ArtifactApi
from .job_run import RunApi
from .model import ModelApi
from .model_version import ModelVersionApi
from .rule import RuleApi
from .attachment import AttachmentApi

__all__ = [
    "Client",
    "RuleApi",
    "JobApi",
    "RunApi",
    "ArtifactApi",
    "DatasetApi",
    "DatasetVersionApi",
    "Page",
    "ModelApi",
    "ModelVersionApi",
    "AttachmentApi",
]



from typing import Optional, List, Union
from urllib.parse import urlencode

from vectice.api.dataset import DatasetApi
from vectice.entity import DatasetVersion

from .Page import Page
from .json_object import JsonObject
from .output.dataset_version_output import DatasetVersionOutput
from .output.paged_response import PagedResponse
from .output.property_output import EntityPropertyOutput


class DatasetVersionApi(DatasetApi):
    def __init__(self, project_token: str, dataset_id: int, _token: Optional[str] = None, allow_self_certificate=True):
        super().__init__(project_token=project_token, _token=_token, allow_self_certificate=allow_self_certificate)
        self._dataset_id = dataset_id
        self._dataset_version_path = super().api_base_path + "/" + str(dataset_id) + "/version"

    @property
    def dataset_id(self) -> int:
        return self._dataset_id

    @property
    def api_base_path(self) -> str:
        return self._dataset_version_path

    def list_dataset_versions(self, page_index=Page.index, page_size=Page.size) -> PagedResponse[DatasetVersionOutput]:
        queries = {"index": page_index, "size": page_size}
        dataset_versions = self._get(self.api_base_path + "?" + urlencode(queries))
        return PagedResponse(
            item_cls=DatasetVersionOutput,
            total=dataset_versions["total"],
            page=dataset_versions["page"],
            items=dataset_versions["items"],
        )

    def get_dataset_version(
        self, search: str, page_index=Page.index, page_size=Page.size
    ) -> Optional[DatasetVersionOutput]:
        queries = {"index": page_index, "size": page_size, "search": search}
        try:
            dataset_version = self._get(self.api_base_path + "?" + urlencode(queries)).get("items")[0]  # type: ignore
        except IndexError:
            return None
        return DatasetVersionOutput(**dataset_version)

    def create_dataset_version(self, dataset_version: JsonObject) -> DatasetVersion:
        return DatasetVersion(self._post(self.api_base_path, dataset_version))

    def update_dataset_version(self, dataset_id: int, dataset_version) -> DatasetVersion:
        return DatasetVersion(self._put(self.api_base_path + "/" + str(dataset_id), dataset_version))

    def create_dataset_version_properties(self, dataset_version_id: int, properties: Union[List[dict], dict]):
        if isinstance(properties, dict):
            DatasetVersion(self._post(self.api_base_path + f"/{dataset_version_id}/entityProperty/", properties))
        elif len(properties) >= 1:
            for prop in properties:
                DatasetVersion(self._post(self.api_base_path + f"/{dataset_version_id}/entityProperty/", prop))

    def update_dataset_version_properties(self, dataset_version_id: int, property_id: int, properties: JsonObject):
        return DatasetVersion(
            self._put(self.api_base_path + f"/{dataset_version_id}/entityProperty/{property_id}", properties)
        )

    def list_dataset_version_properties(
        self, dataset_version_id: int, page_index=Page.index, page_size=Page.size
    ) -> PagedResponse[EntityPropertyOutput]:
        queries = {"index": page_index, "size": page_size}
        model_version_properties = self._get(
            self.api_base_path + f"/{dataset_version_id}/entityProperty?" + urlencode(queries)
        )
        properties = [
            EntityPropertyOutput.from_dict(property).as_dict() for property in model_version_properties["items"]
        ]
        return PagedResponse(
            item_cls=EntityPropertyOutput,
            total=model_version_properties["total"],
            page=model_version_properties["page"],
            items=properties,
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from typing import List, TypeVar, Optional, Tuple

from .property import Property


class WithPropertiesTrait(ABC):
    T = TypeVar("T", bound="WithPropertiesTrait")

    @abstractmethod
    def with_property(self: T, key: str, value: str, timestamp: Optional[datetime] = None) -> T:
        """

        :param key:
        :param value:
        :param timestamp:
        :return: itself
        """
        pass

    @abstractmethod
    def with_extended_properties(self: T, properties: List[Property]) -> T:
        """

        :param properties:
        :return: itself
        """
        pass

    @abstractmethod
    def with_properties(self: T, properties: List[Tuple[str, str]]) -> T:
        """

        :param properties:
        :return: itself
        """
        pass


@dataclass
class WithProperties(WithPropertiesTrait):
    properties: Optional[List[Property]] = None

    T = TypeVar("T", bound="WithProperties")

    def with_property(self: T, key: str, value: str, timestamp: Optional[datetime] = None) -> T:
        self._check_empty_property(key, value)
        if self.properties is None:
            self.properties = []
        if timestamp is None:
            timestamp = datetime.now()
        self.properties.append(Property(key, value, timestamp))
        return self

    def with_extended_properties(self: T, properties: List[Property]) -> T:
        if self.properties is None:
            self.properties = []
        self.properties.extend(properties)
        return self

    def with_properties(self: T, properties: List[Tuple[str, str]]) -> T:
        if len(set(properties)) < len(properties):
            raise ValueError("You can not use the same key value pair more than once.")
        if self.properties is None:
            self.properties = []
        for (key, value) in properties:
            self._check_empty_property(key, value)
            self.properties.append(Property(key, value))
        return self

    @staticmethod
    def _check_empty_property(key: str, value: str):
        if key.strip() == "" or value.strip() == "":
            raise ValueError("Property keys and values can't be empty.")


class WithDelegatedProperties(WithPropertiesTrait, ABC):
    T = TypeVar("T", bound="WithDelegatedProperties")

    @abstractmethod
    def _get_delegate(self) -> WithPropertiesTrait:
        pass

    def with_property(self: T, key: str, value: str, timestamp: Optional[datetime] = None) -> T:
        self._get_delegate().with_property(key, value, timestamp)
        return self

    def with_properties(self: T, properties: List[Tuple[str, str]]) -> T:
        self._get_delegate().with_properties(properties)
        return self

    def with_extended_properties(self: T, properties: List[Property]) -> T:
        self._get_delegate().with_extended_properties(properties)
        return self



import os
from datetime import datetime
from typing import Dict, Optional, List

from dotenv import dotenv_values, find_dotenv


def read_nodejs_date(date_as_string: Optional[str]) -> Optional[datetime]:
    if date_as_string is None:
        return None
    return datetime.strptime(date_as_string, "%Y-%m-%dT%H:%M:%S.%f%z")


def calculate_duration(endDate: datetime, startDateNodeJsFormat: str) -> int:
    # format sample : 2021-06-20T11:04:16.249Z
    startDate = read_nodejs_date(startDateNodeJsFormat)
    if startDate is None:
        raise RuntimeError("Invalid date format for value: " + startDateNodeJsFormat)
    duration = endDate - startDate
    return int(duration.total_seconds())


def read_env(*args: str) -> List[Optional[str]]:
    """
    Read configurations in the order: .vectice file > .env file > environment variables.
    As usual, System variables always take precedence over dotenv files
    And as expected, dotenv file take precedence over dotvectice files.
    """
    if len(args) == 0:
        raise ValueError("At least one argument must be provided to read.")
    config: Dict[str, Optional[str]] = {
        **dotenv_values(find_dotenv(".vectice")),
        **dotenv_values(find_dotenv(".env")),
        **os.environ,
    }
    return [config.get(key) for key in args]



import sys
from typing import Optional, Dict
from vectice.api.client import Client


def get_artifact_id(client: Client, artifact_name: str) -> Optional[int]:
    """
    Helper to get the dataset/model ID
    """
    try:
        artifact_type = str(sys._getframe(1).f_code.co_name)
        if "dataset" in artifact_type:
            return int(
                [
                    artifact.id
                    for artifact in client.list_datasets(artifact_name).list
                    if artifact.name == artifact_name
                ][0]
            )
        elif "model" in artifact_type:
            return int(
                [artifact.id for artifact in client.list_models(artifact_name).list if artifact.name == artifact_name][
                    0
                ]
            )
    except Exception as e:
        raise RuntimeError(f"The artifact ID wasn't found due to {e}")
    return None


def get_artifact_version_id(client: Client, artifact_id: int, version_number: int) -> Optional[int]:
    """
    Helper to get a dataset/model version id
    """
    try:
        artifact_type = str(sys._getframe(1).f_code.co_name)
        if "dataset" in artifact_type:
            return int(
                [
                    version
                    for version in client.list_dataset_versions(artifact_id).list
                    if version.versionNumber == version_number
                ][0].id
            )
        elif "model" in artifact_type:
            return int(
                [
                    version
                    for version in client.list_model_versions(artifact_id).list
                    if version.versionNumber == version_number
                ][0].id
            )
    except Exception as e:
        raise RuntimeError(f"The artifact version ID wasn't found due to {e}")
    return None


def update_artifact_properties(
    client: Client, artifact_id: int, artifact_version_id: int, artifact_model: Dict
) -> None:
    """
    Handles updating model version and dataset version properties
    """
    try:
        artifact_type = str(sys._getframe(1).f_code.co_name)
        if "dataset" in artifact_type:
            update_dataset_version_properties(client, artifact_id, artifact_version_id, artifact_model)
        elif "model" in artifact_type:
            update_model_version_properties(client, artifact_id, artifact_version_id, artifact_model)
    except Exception as e:
        raise RuntimeError(f"Updating properties was unsuccessful due to {e}")
    return None


def update_dataset_version_properties(
    client: Client, artifact_id: int, artifact_version_id: int, artifact_model: Dict
) -> None:
    old_properties = {
        old_prop.key: old_prop.id
        for old_prop in client.list_dataset_version_properties(artifact_id, artifact_version_id).list
    }
    for prop in artifact_model["properties"]:
        if prop.get("key") in old_properties.keys():
            prop["id"] = old_properties.get(prop["key"])
            client.update_dataset_version_properties(
                artifact_id, artifact_version_id, property_id=prop["id"], properties=prop
            )
        elif prop.get("key") not in old_properties.keys():
            client.create_dataset_version_properties(artifact_id, artifact_version_id, properties=prop)


# TODO not implemented yet
def update_model_version_properties(
    client: Client, artifact_id: int, artifact_version_id: int, artifact_model: Dict
) -> None:
    old_properties = {
        old_prop.key: old_prop.id
        for old_prop in client.list_model_version_properties(artifact_id, artifact_version_id).list
    }
    for prop in artifact_model["properties"]:
        if prop.get("key") in old_properties.keys():
            prop["id"] = old_properties.get(prop["key"])
            client.update_model_version_properties(  # type: ignore
                artifact_id, artifact_version_id, property_id=prop["id"], properties=prop
            )
        elif prop.get("key") not in old_properties.keys():
            client.create_model_version_properties(artifact_id, artifact_version_id, properties=prop)  # type: ignore



class VecticeError(Exception):
    def __init__(self, method_name: str, root_cause: Exception):
        self.method = method_name
        self.root_cause = root_cause

    def __str__(self):
        return f"Error executing {self.method}: {str(self.root_cause)}"



from __future__ import annotations

from typing import Optional, List, Union
import logging

from .artifact import Artifact
from .artifact_type import ArtifactType
from .dataset_metadata import DatasetMetadata
from .with_properties import WithDelegatedProperties
from .with_version import WithDelegatedVersion


class DatasetMetadataArtifact(Artifact, WithDelegatedProperties, WithDelegatedVersion):
    """ """

    def __init__(self, dataset: DatasetMetadata, description: Optional[str] = None):
        self.artifactType = ArtifactType.DATASET
        self.description = description
        self.dataset: DatasetMetadata = dataset

    @classmethod
    def create_bigquery(cls, uri: str, description: Optional[str] = None) -> Optional[DatasetMetadataArtifact]:
        """ """
        dataset_metadata_artifact = DatasetMetadata.create_bigquery(uri)
        if dataset_metadata_artifact is not None:
            return cls(dataset_metadata_artifact, description)
        else:
            logging.warning("Failed to create metadata based dataset")
            return None

    @classmethod
    def create_gcs(
        cls, uri: Union[str, List[str]], description: Optional[str] = None
    ) -> Optional[DatasetMetadataArtifact]:
        """ """
        dataset_metadata_artifact = DatasetMetadata.create_gcs(uri)
        if dataset_metadata_artifact is not None:
            return cls(dataset_metadata_artifact, description)
        else:
            logging.warning("Failed to create metadata based dataset")
            return None

    def _get_delegate(self) -> DatasetMetadata:
        return self.dataset



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional, List

from vectice.models.data_resource_path_type import DataResourcePathType
from vectice.models.data_resource_schema import DataResourceSchema
from vectice.models.data_resources_path import DataResourcePath


@dataclass
class DataResource:
    name: str
    description: Optional[str] = None
    path: Optional[DataResourcePath] = None
    schema: Optional[DataResourceSchema] = None

    @classmethod
    def create_item_resource(cls, uri: str) -> DataResource:
        # sanity checks
        if not uri.startswith("gs://"):
            raise RuntimeError("The only file system supported is Google Cloud storage")
        # TODO: chek path is an item (file, table)
        path = uri.replace("gs://", "")
        name = uri.split("/")[-1]
        bucket = path.split("/")[0]
        return DataResource(
            name,
            "",
            path=DataResourcePath(name, bucket, path, uri, path, DataResourcePathType.get_file_type(path), False),
            schema=DataResourceSchema(
                DataResourcePathType.get_file_type(path), name, "", DataResourcePathType.get_file_type(path)
            ),
        )

    @classmethod
    def create_folder_resource(cls, folder: str) -> DataResource:
        # sanity checks
        if not folder.startswith("gs://"):
            raise RuntimeError("The only file system supported is Google Cloud storage")
        # TODO: check path is a folder (directory, schema)
        if not folder.endswith("/"):
            folder = folder + "/"
        path = folder.replace("gs://", "")
        name = folder.split("/")[-2]
        bucket = path.split("/")[0]
        return DataResource(
            name,
            "",
            path=DataResourcePath(name, bucket, path, folder, path, DataResourcePathType.Folder, True),
            schema=DataResourceSchema(DataResourcePathType.Folder, name, "", DataResourcePathType.Folder),
        )

    @classmethod
    def create_resources(cls, files: Optional[List[str]], folders: Optional[List[str]] = None):
        result = []
        if files:
            result.extend([DataResource.create_item_resource(file) for file in files])
        if folders:
            result.extend([DataResource.create_folder_resource(folder) for folder in folders])
        return result
